#ifndef CONFIG_H
#define CONFIG_H

#define INSTALL_PREFIX "/usr/local"
#define PLUGIN_PATH "/usr/local/lib/wayfire"
#define PLUGIN_XML_DIR "/usr/local/share/wayfire/metadata"
#define SYSCONFDIR "/etc"
#define WF_DEFAULT_CONFIG_BACKEND "/usr/local/lib/wayfire/libdefault-config-backend.so"
#define BUILD_WITH_IMAGEIO
#define USE_GLES32
#define WF_HAS_XWAYLAND 1


#endif /* end of include guard: CONFIG_H */
